@@ -3,7 +3,7 @@ import calendar
 import os
 import plotly.express as px
 import plotly.graph_objs as go
-from datetime import datetime, timedelta
+from dateutil.parser import parse
 
 class dataprocessing:
     
@@ -32,8 +32,8 @@ class dataprocessing:
 
         print("\n5. plot_two")
         print("   - Description: Plots specified columns from two different DataFrames using a shared date column. Useful for comparing data.")
-        print("   - Usage: plot_two(df1, col1, df2, col2, date_column)")
-        print("   - Example: plot_two(df1, 'cost', df2, 'cost', 'obs')")
+        print("   - Usage: plot_two(df1, col1, df2, col2, date_column, same_axis=True)")
+        print("   - Example: plot_two(df1, 'cost', df2, 'cost', 'obs', True)")
 
         print("\n6. remove_nan_rows")
         print("   - Description: Removes rows from a DataFrame where the specified column has NaN values.")
@@ -50,10 +50,10 @@ class dataprocessing:
         print("   - Usage: plot_one(df1, col1, date_column)")
         print("   - Example: plot_one(df, 'Spend', 'OBS')")
 
-        print("\n9. convert_weekly_to_date")
+        print("\n9. week_of_year_mapping")
         print("   - Description: Converts a week column in 'yyyy-Www' or 'yyyy-ww' format to week commencing date.")
-        print("   - Usage: convert_weekly_to_date(df, column_name, first_week_start_date)")
-        print("   - Example: convert_weekly_to_date(df, 'week', '30/12/2019')")
+        print("   - Usage: week_of_year_mapping(df, week_col, start_day_str)")
+        print("   - Example: week_of_year_mapping(df, 'week', 'mon')")
 
         print("\n10. exclude_rows")
         print("    - Description: Removes rows from a DataFrame based on whether the values in a specified column are not in a provided list.")
@@ -73,14 +73,35 @@ class dataprocessing:
 
         print("\n13. merge_dataframes_on_column")
         print("    - Description: Merge a list of DataFrames on a common column.")
-        print("    - Usage: merge_dataframes_on_column(dataframes, common_column='OBS', merge_how='inner')")
-        print("    - Example: merge_dataframes_on_column([df1, df2, df3], common_column='OBS', merge_how='inner')")
+        print("    - Usage: merge_dataframes_on_column(dataframes, common_column='OBS', merge_how='outer')")
+        print("    - Example: merge_dataframes_on_column([df1, df2, df3], common_column='OBS', merge_how='outer')")
 
         print("\n14. merge_and_update_dfs")
         print("    - Description: Merges two dataframes on a key column, updates the first dataframe's columns with the second's where available, and returns a dataframe sorted by the key column.")
         print("    - Usage: merge_and_update_dfs(df1, df2, key_column)")
         print("    - Example: merged_dataframe = merge_and_update_dfs(processed_facebook, finalised_meta, 'OBS')")
 
+        print("\n15. convert_us_to_uk_dates")
+        print("    - Description: Convert a DataFrame column with mixed date formats to datetime.")
+        print("    - Usage: convert_us_to_uk_dates(df, date_col)")
+        print("    - Example: convert_us_to_uk_dates(df, 'date')")
+        
+        print("\n16. combine_sheets")
+        print("    - Description: Combines multiple DataFrames from a dictionary into a single DataFrame.")
+        print("    - Usage: combine_sheets(all_sheets)")
+        print("    - Example: combine_sheets({'Sheet1': df1, 'Sheet2': df2})")
+        
+        print("\n17. dynamic_pivot")
+        print("    - Description: Dynamically pivots a DataFrame based on specified columns.")
+        print("    - Usage: dynamic_pivot(data_frame, index_col, columns, values_col, fill_value=0)")
+        print("    - Example: dynamic_pivot(df, 'Date', ['Category1', 'Category2'], ['Value1'])")
+        
+        print("\n18. classify_within_column")
+        print("    - Description: Allows you to map a dictionary of substrings within a column.")
+        print("    - Usage: classify_within_column(df, col_name, to_find_dict, if_not_in_country_dict='Other')")
+        print("    - Example: classify_within_column(df, 'campaign', {'uk_': 'uk'}, 'other')")
+
+
     def get_wd_levels(self, levels):
         """
         Gets the current wd of whoever is working on it and gives the options to move the number of levels up.
@@ -135,11 +156,11 @@ class dataprocessing:
 
         return modified_data_frame
     
-    def aggregate_daily_to_wc(self, df, date_column, group_columns, sum_columns, wc, aggregation='sum', include_totals=False):
+    def aggregate_daily_to_wc_long(self, df : pd.DataFrame, date_column : str, group_columns : list[str], sum_columns : list[str], wc : str = 'sun', aggregation : str = 'sum'):
         """
         Aggregates daily data into weekly data, starting on a specified day of the week, 
         and groups the data by additional specified columns. It aggregates specified numeric columns 
-        by summing or averaging them, and pivots the data to create separate columns for each combination 
+        by summing, averaging, or counting them, and pivots the data to create separate columns for each combination 
         of the group columns and sum columns. NaN values are replaced with 0 and the index is reset. 
         The day column is renamed from 'Day' to 'OBS'.
 
@@ -155,9 +176,7 @@ class dataprocessing:
         - wc: string
             The week commencing day (e.g., 'sun' for Sunday, 'mon' for Monday).
         - aggregation: string, optional (default 'sum')
-            Aggregation method, either 'sum' or 'average'.
-        - include_totals: boolean, optional (default False)
-            If True, include total columns for each sum_column.
+            Aggregation method, either 'sum', 'average', or 'count'.
 
         Returns:
         - pandas DataFrame
@@ -190,12 +209,57 @@ class dataprocessing:
         # Group by the new week start column and additional columns, then aggregate the numeric columns
         if aggregation == 'average':
             grouped = df_copy.groupby(['week_start'] + group_columns)[sum_columns].mean().reset_index()
+        elif aggregation == 'count':
+            grouped = df_copy.groupby(['week_start'] + group_columns)[sum_columns].count().reset_index()
         else:  # Default to 'sum' if any other value is provided
             grouped = df_copy.groupby(['week_start'] + group_columns)[sum_columns].sum().reset_index()
 
         # Rename 'week_start' column to 'OBS'
         grouped = grouped.rename(columns={'week_start': 'OBS'})
 
+        # Reset the index of the final DataFrame
+        grouped.reset_index(inplace = True)
+
+        return grouped
+    
+    def aggregate_daily_to_wc_wide(self, df : pd.DataFrame, date_column : str, group_columns : list[str], sum_columns : list[str], wc : str = 'sun', aggregation : str = 'sum', 
+                                   include_totals : bool = False):
+        
+                """
+        Aggregates daily data into weekly data, starting on a specified day of the week, 
+        and groups the data by additional specified columns. It aggregates specified numeric columns 
+        by summing, averaging, or counting them, and pivots the data to create separate columns for each combination 
+        of the group columns and sum columns. NaN values are replaced with 0 and the index is reset. 
+        The day column is renamed from 'Day' to 'OBS'.
+
+        Parameters:
+        - df: pandas DataFrame
+            The input DataFrame containing daily data.
+        - date_column: string
+            The name of the column in the DataFrame that contains date information.
+        - group_columns: list of strings
+            Additional column names to group by along with the weekly grouping.
+        - sum_columns: list of strings
+            Numeric column names to be aggregated during aggregation.
+        - wc: string
+            The week commencing day (e.g., 'sun' for Sunday, 'mon' for Monday).
+        - aggregation: string, optional (default 'sum')
+            Aggregation method, either 'sum', 'average', or 'count'.
+        - include_totals: boolean, optional (default False)
+            If True, include total columns for each sum_column.
+
+
+
+        Returns:
+        - pandas DataFrame
+            A new DataFrame with weekly aggregated data. The index is reset,
+            and columns represent the grouped and aggregated metrics. The DataFrame 
+            is in wide format, with separate columns for each combination of 
+            grouped metrics.
+        """
+        
+        grouped = self.aggregate_daily_to_wc_long(df, date_column, group_columns, sum_columns, wc, aggregation)
+        
         # Pivot the data to wide format
         if group_columns:
             wide_df = grouped.pivot_table(index='OBS', 
@@ -203,7 +267,7 @@ class dataprocessing:
                                         values=sum_columns,
                                         aggfunc='first')
             # Flatten the multi-level column index and create combined column names
-            wide_df.columns = [' '.join(col).strip() for col in wide_df.columns.values]
+            wide_df.columns = ['_'.join(col).strip() for col in wide_df.columns.values]
         else:
             wide_df = grouped.set_index('OBS')
 
@@ -223,8 +287,8 @@ class dataprocessing:
         # Reset the index of the final DataFrame
         wide_df = wide_df.reset_index()
 
-        return wide_df      
-  
+        return wide_df
+    
     def convert_monthly_to_daily(self, df, date_column):
         """
         Convert a DataFrame with monthly data to daily data.
@@ -346,31 +410,34 @@ class dataprocessing:
 
         return fig
 
-    def convert_weekly_to_date(self, df, column_name, first_week_start_date):
-        def get_week_start_date(week_code):
-            if '-W' in week_code:
-                # Handle ISO week date format
-                year, week_num = map(int, week_code.split('-W'))
-                year_start = datetime(year, 1, 1)
-                first_monday = year_start - timedelta(days=year_start.isoweekday() - 1)
-                week_start_date = first_monday + timedelta(weeks=week_num - 1)
-            else:
-                # Handle regular week number format
-                year, week_num = map(int, week_code.split('-'))
-                first_week_start = datetime.strptime(first_week_start_date, '%d/%m/%Y')
-                # Adjust for the year
-                first_week_start = first_week_start.replace(year=year)
-                week_start_date = first_week_start + timedelta(weeks=week_num - 1)
-            
-            return week_start_date.strftime('%d/%m/%Y')
-        
-        # Create a mapping of week codes to dates
-        unique_weeks = df[column_name].unique()
-        week_to_date_mapping = {week: get_week_start_date(week) for week in unique_weeks}
+    def week_of_year_mapping(self, df, week_col, start_day_str):
+        from datetime import datetime, timedelta
+        # Mapping of string day names to day numbers (1 for Monday, 7 for Sunday)
+        day_mapping = {
+            'mon': 1, 'tue': 2, 'wed': 3, 'thu': 4, 'fri': 5, 'sat': 6, 'sun': 7
+        }
 
-        # Map each row in the DataFrame to the corresponding date
-        df[f'{column_name}_start_date'] = df[column_name].map(week_to_date_mapping)
+        # Convert the day string to a number, or raise an error if not valid
+        start_day = day_mapping.get(start_day_str.lower())
+        if start_day is None:
+            raise ValueError(f"Invalid day input: '{start_day_str}'. Please use one of 'mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun'.")
 
+        # Function to convert week number to start date of the week
+        def week_to_startdate(week_str, start_day):
+            year, week = map(int, week_str.split('-W'))
+            first_day_of_year = datetime(year, 1, 1)
+            day_of_week = first_day_of_year.isocalendar()[2]
+            days_to_add = (7 - day_of_week + 1) if day_of_week > 4 else (1 - day_of_week)
+            start_of_iso_week = first_day_of_year + timedelta(days=days_to_add)
+
+            # Adjust start day
+            days_to_shift = (start_day - 1) % 7
+            start_of_week = start_of_iso_week + timedelta(days=days_to_shift)
+
+            return start_of_week + timedelta(weeks=week - 1)
+
+        # Apply the function to each row in the specified week column
+        df['OBS'] = df[week_col].apply(lambda x: week_to_startdate(x, start_day)).dt.strftime('%d/%m/%Y')
         return df
     
     def exclude_rows(self, df, col_to_filter, list_of_filters):
@@ -437,7 +504,7 @@ class dataprocessing:
 
         return merged_df
     
-    def merge_dataframes_on_column(self, dataframes, common_column='OBS', merge_how='inner'):
+    def merge_dataframes_on_column(self, dataframes, common_column='OBS', merge_how='outer'):
         """
         Merge a list of DataFrames on a common column.
 
@@ -457,8 +524,12 @@ class dataprocessing:
         for df in dataframes[1:]:
             merged_df = pd.merge(merged_df, df, on=common_column, how=merge_how)
 
+        merged_df[common_column] = pd.to_datetime(merged_df[common_column])
+        merged_df = merged_df.sort_values(by=common_column)
+        merged_df = merged_df.fillna(0)
+        
         return merged_df
-
+    
     def merge_and_update_dfs(self, df1, df2, key_column):
         """
         Merges two dataframes on a key column, updates the first dataframe's columns with the second's where available,
@@ -493,4 +564,111 @@ class dataprocessing:
         # Handle null values (optional, can be adjusted as needed)
         merged_df.fillna(0, inplace=True)
 
-        return merged_df
\ No newline at end of file
+        return merged_df
+    
+    def convert_us_to_uk_dates(self, df, date_col):
+        import datetime
+        def fix_date(d):
+            # Convert datetime objects to string
+            if isinstance(d, pd.Timestamp) or isinstance(d, datetime.datetime):
+                d = d.strftime('%m/%d/%Y')
+            
+            # Split date string into components
+            parts = d.split('/')
+            
+            # Check for two formats: mm/dd/yyyy and dd/mm/yyyy
+            if len(parts) == 3:
+                year, month, day = parts[2], parts[0], parts[1]
+                # Correct for two-digit years
+                if len(year) == 2:
+                    year = '20' + year
+                # Identify dates needing correction: where month > 12 or it follows the yyyy-dd-mm pattern
+                if int(month) > 12 or (int(day) <= 12 and int(year) > 31):
+                    # Assume year is correct, flip 'month' and 'day' for correction
+                    month, day = day, month
+                return f"{year}-{month.zfill(2)}-{day.zfill(2)}"
+            else:
+                # Handle already correct or non-standard formats cautiously
+                return d
+
+        # Apply the fix to the specified column
+        df[date_col] = df[date_col].apply(lambda x: fix_date(x) if not pd.isnull(x) else x)
+        return df
+
+    def combine_sheets(self, all_sheets):
+        """
+        Combines multiple DataFrames from a dictionary into a single DataFrame.
+        Adds a column 'SheetName' indicating the origin sheet of each row.
+
+        Parameters:
+        all_sheets (dict): A dictionary of DataFrames, typically read from an Excel file with multiple sheets.
+
+        Returns:
+        DataFrame: A concatenated DataFrame with an additional 'SheetName' column.
+        """
+        combined_df = pd.DataFrame()
+
+        for sheet_name, df in all_sheets.items():
+            df['SheetName'] = sheet_name 
+            combined_df = pd.concat([combined_df, df], ignore_index=True)
+
+        return combined_df
+    
+    def dynamic_pivot(self, data_frame, index_col, columns, values_col, fill_value=0):
+        # Ensure OBS is in datetime format for proper sorting
+        data_frame[index_col] = pd.to_datetime(data_frame[index_col], dayfirst=True)
+        
+        # Check if values_col is a single column or a list and pivot accordingly
+        if isinstance(values_col, list):
+            # If values_col is a list, use .pivot_table() to accommodate multiple values columns
+            pivoted_df = data_frame.pivot_table(index=index_col, columns=columns, values=values_col, aggfunc='sum')
+        else:
+            # For a single value column, use .pivot()
+            pivoted_df = data_frame.pivot(index=index_col, columns=columns, values=values_col)
+        
+        # Handling MultiIndex columns if present, making them a flat structure
+        if isinstance(pivoted_df.columns, pd.MultiIndex):
+            pivoted_df.columns = ['_'.join(map(str, col)).strip() for col in pivoted_df.columns.values]
+        else:
+            pivoted_df.columns = pivoted_df.columns.map(str)
+        
+        # Reset the pivot before returning
+        pivoted_df = pivoted_df.reset_index()
+        
+        # Sort by OBS from oldest to newest
+        pivoted_df[index_col] = pd.to_datetime(pivoted_df[index_col])  # Ensure sorting works correctly
+        pivoted_df = pivoted_df.sort_values(by=index_col)
+        
+        # Convert OBS back to a string in YYYY-MM-DD format for display purposes
+        pivoted_df[index_col] = pivoted_df[index_col].dt.strftime('%Y-%m-%d')
+        
+        # Fill in any NaNs
+        pivoted_df = pivoted_df.fillna(fill_value)
+        
+        return pivoted_df
+    
+    def classify_within_column(self, df, col_name, to_find_dict, if_not_in_country_dict="Other"):
+        """
+        Classify entries in a DataFrame column based on a dictionary of substrings to class mappings.
+        
+        Parameters:
+        - df: pandas.DataFrame to operate on.
+        - col_name: String, name of the column to classify.
+        - to_find_dict: Dictionary, where keys are substrings to find and values are the corresponding classifications.
+        - if_not_in_country_dict: String, default classification if no substring matches are found.
+        
+        Returns:
+        - DataFrame with a new column '<col_name>_mapping' containing the classification results.
+        """
+        # Define the inner function for classification
+        def find_word_in_string(x_string, to_find_dict=to_find_dict, default_value=if_not_in_country_dict):
+            x_string_lower = x_string.lower()
+            for key, value in to_find_dict.items():
+                if key.lower() in x_string_lower:
+                    return value
+            return default_value
+
+        # Apply the inner function to the specified column and create a new column with the results
+        df[col_name + '_mapping'] = df[col_name].apply(find_word_in_string)
+        
+        return df
